//--------------------------------------------------------------------------------------
// Order Independent Transparency with Dual Depth Peeling
//
// Author: Louis Bavoil
// Email: sdkfeedback@nvidia.com
//
// Depth peeling is traditionally used to perform order independent transparency (OIT)
// with N geometry passes for N transparency layers. Dual depth peeling enables peeling
// N transparency layers in N/2+1 passes, by peeling from the front and the back
// simultaneously using a min-max depth buffer. This sample performs either normal or
// dual depth peeling and blends on the fly.
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

unit DepthPeeling;

interface
uses
  dglopengl, ShaderClass;

type
  TDrawVertex = procedure of object;

const
  g_numPasses = 4;
  MAX_DEPTH = 1.0;

var
  g_shaderDualInit,
  g_shaderDualPeel_Color,
  g_shaderDualPeel_ColorTextur,
  g_shaderDualPeel_Textur,
  g_shaderDualBlend,
  g_shaderDualFinal: TShader;

  g_dualDepthTexId: array [0..1] of GLuint;
  g_dualFrontBlenderTexId: array [0..1] of GLuint;
  g_dualBackTempTexId: array [0..1] of GLuint;
  g_dualPeelingSingleFboId: GLuint;
  g_dualBackBlenderTexId: GLuint;
  g_dualBackBlenderFboId: GLuint;

  g_quadDisplayList: GLuint;

  g_imageWidth,
  g_imageHeight: Integer;

  g_queryId: GLuint;

  g_drawBuffers: array [0..6] of GLenum = (
              GL_COLOR_ATTACHMENT0,
						  GL_COLOR_ATTACHMENT1,
						  GL_COLOR_ATTACHMENT2,
						  GL_COLOR_ATTACHMENT3,
						  GL_COLOR_ATTACHMENT4,
						  GL_COLOR_ATTACHMENT5,
						  GL_COLOR_ATTACHMENT6);

  g_backgroundColor: TGLVectorf3;

  g_Counter: Extended;

procedure InitGL;
procedure Render;
procedure RenderDualPeeling(ShaderMode: Integer);
procedure ReloadShaders;
procedure reshape;
implementation

uses
  Classes, Windows, Dialogs, SysUtils, Figures;

procedure CHECK_GL_ERRORS(Pos: string);
var
  err: GLenum;
begin
  Pos := Pos + ' ';
  err := glGetError;
  if (err <> GL_NO_ERROR) then
  case err of
    GL_INVALID_ENUM:      raise Exception.Create(Pos + 'GL_INVALID_ENUM');
    GL_INVALID_VALUE:     raise Exception.Create(Pos + 'GL_INVALID_VALUE');
    GL_INVALID_OPERATION: raise Exception.Create(Pos + 'GL_INVALID_OPERATION');
    GL_OUT_OF_MEMORY:     raise Exception.Create(Pos + 'GL_OUT_OF_MEMORY');
    else
                          raise Exception.Create('other Error');
  end;

  err := glCheckFramebufferStatus(GL_FRAMEBUFFER);
  case err of
    GL_FRAMEBUFFER_COMPLETE: ;
    GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:         raise Exception.Create(Pos + 'Incomplete attachment');
    GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: raise Exception.Create(Pos + 'Missing attachment');
    GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:     raise Exception.Create(Pos + 'Incomplete dimensions');
    GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:        raise Exception.Create(Pos + 'Incomplete formats');
    GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:    raise Exception.Create(Pos + 'Incomplete draw buffer');
    GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:    raise Exception.Create(Pos + 'Incomplete read buffer');
    GL_FRAMEBUFFER_UNSUPPORTED_EXT:               raise Exception.Create(Pos + 'Framebufferobjects unsupported');
  end;
end;

procedure DeleteDualPeelingRenderTargets;
begin
	glDeleteFramebuffers(1, @g_dualBackBlenderFboId);
	glDeleteFramebuffers(1, @g_dualPeelingSingleFboId);
	glDeleteTextures(2, @g_dualDepthTexId);
	glDeleteTextures(2, @g_dualFrontBlenderTexId);
	glDeleteTextures(2, @g_dualBackTempTexId);
	glDeleteTextures(1, @g_dualBackBlenderTexId);
end;

procedure BuildShaders;
begin
  g_shaderDualInit := TShader.Create('dual_peeling_init_vertex.glsl', 'dual_peeling_init_fragment.glsl');

  g_shaderDualPeel_Color := TShader.Create;
  g_shaderDualPeel_Color.AddShaderFile(GL_VERTEX_SHADER, 'dual_peeling_peel_vertex.glsl');
  g_shaderDualPeel_Color.AddShaderFile(GL_FRAGMENT_SHADER, 'shade_fragment color.glsl');
  g_shaderDualPeel_Color.AddShaderFile(GL_FRAGMENT_SHADER, 'dual_peeling_peel_fragment.glsl');
  g_shaderDualPeel_Color.CompileAndLink;

  g_shaderDualPeel_Textur := TShader.Create;
  g_shaderDualPeel_Textur.AddShaderFile(GL_VERTEX_SHADER, 'dual_peeling_peel_vertex.glsl');
  g_shaderDualPeel_Textur.AddShaderFile(GL_FRAGMENT_SHADER, 'shade_fragment textur.glsl');
  g_shaderDualPeel_Textur.AddShaderFile(GL_FRAGMENT_SHADER, 'dual_peeling_peel_fragment.glsl');
  g_shaderDualPeel_Textur.CompileAndLink;

  g_shaderDualPeel_ColorTextur := TShader.Create;
  g_shaderDualPeel_ColorTextur.AddShaderFile(GL_VERTEX_SHADER, 'dual_peeling_peel_vertex.glsl');
  g_shaderDualPeel_ColorTextur.AddShaderFile(GL_FRAGMENT_SHADER, 'shade_fragment color_textur.glsl');
  g_shaderDualPeel_ColorTextur.AddShaderFile(GL_FRAGMENT_SHADER, 'dual_peeling_peel_fragment.glsl');
  g_shaderDualPeel_ColorTextur.CompileAndLink;

  g_shaderDualBlend := TShader.Create('dual_peeling_blend_vertex.glsl', 'dual_peeling_blend_fragment.glsl');
  g_shaderDualFinal := TShader.Create('dual_peeling_final_vertex.glsl', 'dual_peeling_final_fragment.glsl');
end;

procedure DestroyShaders;
begin
	g_shaderDualInit.Free;

	g_shaderDualPeel_Color.Free;
  g_shaderDualPeel_Textur.Free;
  g_shaderDualPeel_ColorTextur.Free;

	g_shaderDualBlend.Free;
	g_shaderDualFinal.Free;
end;

procedure ReloadShaders;
begin
	DestroyShaders;
	BuildShaders;
end;

procedure InitDualPeelingRenderTargets;
var
  i, j: Integer;
begin
  glGenTextures(2, @g_dualDepthTexId);
	glGenTextures(2, @g_dualFrontBlenderTexId);
	glGenTextures(2, @g_dualBackTempTexId);
	glGenFramebuffers(1, @g_dualPeelingSingleFboId);

  for i := 0 to 1 do
 	begin
		glBindTexture(GL_TEXTURE_RECTANGLE, g_dualDepthTexId[i]);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RG32F{GL_FLOAT_RG32_NV},
      g_imageWidth, g_imageHeight, 0, GL_RGB, GL_FLOAT, 0);
    // Wenn GL_FLOAT_RG32_NV genutzt wird, dann muss in dual_peeling_peel_fragment.glsl
    // vec2 depthBlender = texture2DRect(DepthBlenderTex, gl_FragCoord.xy).xw; -> .. gl_FragCoord.xy).xy;
    // gendert werden;


		glBindTexture(GL_TEXTURE_RECTANGLE, g_dualFrontBlenderTexId[i]);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA, g_imageWidth, g_imageHeight,
					 0, GL_RGBA, GL_FLOAT, 0);

		glBindTexture(GL_TEXTURE_RECTANGLE, g_dualBackTempTexId[i]);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA, g_imageWidth, g_imageHeight,
					 0, GL_RGBA, GL_FLOAT, 0);
 	end;

	glGenTextures(1, @g_dualBackBlenderTexId);
	glBindTexture(GL_TEXTURE_RECTANGLE, g_dualBackBlenderTexId);
	glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGB, g_imageWidth, g_imageHeight, 0, GL_RGB, GL_FLOAT, 0);

	glGenFramebuffers(1, @g_dualBackBlenderFboId);
	glBindFramebuffer(GL_FRAMEBUFFER, g_dualBackBlenderFboId);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_RECTANGLE, g_dualBackBlenderTexId, 0);

	glBindFramebuffer(GL_FRAMEBUFFER, g_dualPeelingSingleFboId);
	j := 0;
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_RECTANGLE, g_dualDepthTexId[j], 0);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_RECTANGLE, g_dualFrontBlenderTexId[j], 0);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_RECTANGLE, g_dualBackTempTexId[j], 0);
	j := 1;
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT3, GL_TEXTURE_RECTANGLE, g_dualDepthTexId[j], 0);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT4, GL_TEXTURE_RECTANGLE, g_dualFrontBlenderTexId[j], 0);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT5, GL_TEXTURE_RECTANGLE, g_dualBackTempTexId[j], 0);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT6, GL_TEXTURE_RECTANGLE, g_dualBackBlenderTexId, 0);

	CHECK_GL_ERRORS('Init targets');
end;

procedure RenderDDP_Init;
begin
  // ---------------------------------------------------------------------
  // 1. Initialize Min-Max Depth Buffer
  // ---------------------------------------------------------------------
  glBindFramebuffer(GL_FRAMEBUFFER, g_dualPeelingSingleFboId);
  // Render targets 1 and 2 store the front and back colors
  // Clear to 0.0 and use MAX blending to filter written color
  // At most one front color and one back color can be written every pass
  glDrawBuffers(2, @g_drawBuffers[1]);
  glClearColor(0, 0, 0, 0);
  glClear(GL_COLOR_BUFFER_BIT);
  // Render target 0 stores (-minDepth, maxDepth, alphaMultiplier)
  glDrawBuffer(g_drawBuffers[0]);
  glClearColor(-MAX_DEPTH, -MAX_DEPTH, 0, 0);
  glClear(GL_COLOR_BUFFER_BIT);
  glBlendEquation(GL_MAX);
  g_shaderDualInit.UseShader;
  Render;
  g_shaderDualInit.UnUseShader;
  CHECK_GL_ERRORS('Render Init');
end;

procedure RenderDDP_Final(currId: Integer);
begin
  // ---------------------------------------------------------------------
  // 3. Final Pass
  // ---------------------------------------------------------------------
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
  glDrawBuffer(GL_BACK);
  g_shaderDualFinal.UseShader;
  g_shaderDualFinal.bindTextureRECT('DepthBlenderTex', g_dualDepthTexId[currId], 1);
  g_shaderDualFinal.bindTextureRECT('FrontBlenderTex', g_dualFrontBlenderTexId[currId], 2);
  g_shaderDualFinal.bindTextureRECT('BackBlenderTex', g_dualBackBlenderTexId, 3);
  glCallList(g_quadDisplayList);
  g_shaderDualFinal.UnUseShader;
  CHECK_GL_ERRORS('Render Final');
end;

procedure RenderDualPeeling(ShaderMode: Integer);
var
  pass,
  currId, prevId, bufId: Integer;
  sample_count: GLuint;
  opacity: GLfloat;
  ShaderPeel: TShader;
begin
 	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

  RenderDDP_Init;

	// ---------------------------------------------------------------------
	// 2. Dual Depth Peeling + Blending
	// ---------------------------------------------------------------------

  case ShaderMode of
    0: ShaderPeel := g_shaderDualPeel_Color;
    1: ShaderPeel := g_shaderDualPeel_Textur;
    2: ShaderPeel := g_shaderDualPeel_ColorTextur;
  end;

	glDrawBuffer(g_drawBuffers[6]);
	glClearColor(g_backgroundColor[0], g_backgroundColor[1], g_backgroundColor[2], 0.0);
	glClear(GL_COLOR_BUFFER_BIT);

  for pass := 1 to g_numPasses - 1 do
  begin
		currId := pass mod 2;
		prevId := 1 - currId;
		bufId := currId * 3;

		glDrawBuffers(2, @g_drawBuffers[bufId+1]);
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT);

		glDrawBuffer(g_drawBuffers[bufId+0]);
		glClearColor(-MAX_DEPTH, -MAX_DEPTH, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT);

		// Render target 0: RG32F MAX blending
		// Render target 1: RGBA MAX blending
		// Render target 2: RGBA MAX blending
		glDrawBuffers(3, @g_drawBuffers[bufId+0]);
		glBlendEquation(GL_MAX);

    ShaderPeel.UseShader;
		ShaderPeel.bindTextureRECT('DepthBlenderTex', g_dualDepthTexId[prevId], 1);
		ShaderPeel.bindTextureRECT('FrontBlenderTex', g_dualFrontBlenderTexId[prevId], 2);
  		Render;
    ShaderPeel.UnUseShader;

		CHECK_GL_ERRORS('Render Main 1');

		// Full screen pass to alpha-blend the back color
		glDrawBuffer(g_drawBuffers[6]);

		glBlendEquation(GL_FUNC_ADD);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    g_shaderDualBlend.UseShader;
		g_shaderDualBlend.bindTextureRECT('TempTex', g_dualBackTempTexId[currId], 1);
		glCallList(g_quadDisplayList);
    g_shaderDualBlend.UnUseShader;

		CHECK_GL_ERRORS('Render Main 2');
  end;

	glDisable(GL_BLEND);

  RenderDDP_Final(currId);
end;

procedure MakeFullScreenQuad;
begin
	g_quadDisplayList := glGenLists(1);
	glNewList(g_quadDisplayList, GL_COMPILE);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);
	glBegin(GL_QUADS);
		glVertex2f(0.0, 0.0);
		glVertex2f(1.0, 0.0);
		glVertex2f(1.0, 1.0);
		glVertex2f(0.0, 1.0);
	glEnd();
	glPopMatrix();

	glEndList();
end;

procedure reshape;
var
  viewport: TGLVectori4;
begin
  if Assigned(g_shaderDualInit) then
  begin
    glGetIntegerv(GL_VIEWPORT, @viewport );
    g_imageWidth := viewport[2] - viewport[0];
    g_imageHeight := viewport[3] - viewport[1];

    DeleteDualPeelingRenderTargets();
    InitDualPeelingRenderTargets();
  end;
end;

procedure Render;
begin
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;

  glTranslatef(0, 0, -5);
  glColor4f(1, 0, 0, 0.3);
  DrawBoxObject(0.75, 0.75, 0.75, false, voffset, toffset, tscale);

  glPushMatrix;
    glColor4f(0, 1, 0, 0.3);
    glRotatef(g_Counter, 0, 1, 0);
    glTranslatef(-0.7, 0, 0);
    DrawBoxObject(0.75, 0.75, 0.75, false, voffset, toffset, tscale);
  glPopMatrix;
end;

procedure InitGL;
var
  viewport: TGLVectori4;
  i: GLint;
begin
  glGetIntegerv(GL_MAX_TEXTURE_UNITS, @i);
  if i < 4 then
    Halt;

  glGetIntegerv(GL_VIEWPORT, @viewport );
  g_imageWidth := viewport[2] - viewport[0];
  g_imageHeight := viewport[3] - viewport[1];

  InitDualPeelingRenderTargets;
  glBindFramebuffer(GL_FRAMEBUFFER, 0);

	BuildShaders;
	MakeFullScreenQuad;

	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glDisable(GL_NORMALIZE);

	glGenQueries(1, @g_queryId);
end;

end.
